import os
import pickle
import subprocess
import sys
from argparse import Namespace
from copy import deepcopy
from csv import DictReader
from tempfile import TemporaryDirectory
from time import time
from typing import List

from .automatix import Automatix
from .command import Command, SkipBatchItemException, AbortException
from .config import (
    arguments, CONFIG, get_script, LOG, update_script_from_row, collect_vars, SCRIPT_FIELDS, VERSION, init_logger,
    PROGRESS_BAR, progress_bar
)
from .parallel import screen_switch_loop


def check_for_original_automatix():
    p = subprocess.run('pip list | grep automatix', shell=True, stdout=subprocess.PIPE)

    for line in p.stdout.decode().split('\n'):
        if line.strip().split(' ', maxsplit=1)[0].strip() == 'automatix':
            raise Exception(
                'This package MUST NOT be installed along with the "automatix" package.'
                ' Both packages use the same entry point and module names and therefore'
                ' are conflicting. Please uninstall automatix AND automatix_cmd first,'
                ' THEN reinstall the package you want to use!')


def setup(args: Namespace):
    """Setup logger and print version information"""
    init_logger(name=CONFIG['logger'], debug=args.debug)

    LOG.info(f'Automatix Version {VERSION} (automatix_cmd)')

    configfile = CONFIG.get('config_file')
    if configfile:
        LOG.info(f'Using configuration from: {configfile}')
    else:
        LOG.warning('Configuration file not found or not configured. Using defaults.')


def get_command_class() -> type:
    if CONFIG.get('bundlewrap'):
        from .bundlewrap import BWCommand, AutomatixBwRepo

        CONFIG['bw_repo'] = AutomatixBwRepo(repo_path=os.environ.get('BW_REPO_PATH', '.'))
        return BWCommand
    else:
        return Command


def get_script_and_batch_items(args: Namespace) -> (dict, list):
    script = get_script(args=args)

    batch_items: List[dict] = [{}]
    if args.vars_file:
        with open(args.vars_file) as csvfile:
            batch_items = list(DictReader(csvfile))
        script['batch_mode'] = False if args.parallel else True
        script['batch_items_count'] = 1 if args.parallel else len(batch_items)
        LOG.notice(f'Detected {"parallel" if args.parallel else "batch"} processing from CSV file.')

    if args.steps:
        exclude = script['exclude'] = args.steps.startswith('e')
        script['steps'] = {int(s) for s in (args.steps[1:] if exclude else args.steps).split(',')}

    return script, batch_items


def run_batch_items(script: dict, batch_items: list, args: Namespace):
    cmd_class = get_command_class()

    for i, row in enumerate(batch_items, start=1):
        script_copy = deepcopy(script)
        update_script_from_row(row=row, script=script_copy, index=i)

        variables = collect_vars(script_copy)

        auto = Automatix(
            script=script_copy,
            variables=variables,
            config=CONFIG,
            cmd_class=cmd_class,
            script_fields=SCRIPT_FIELDS,
            cmd_args=args,
            batch_index=i,
        )
        auto.env.attach_logger()
        auto.set_command_count()
        try:
            auto.run()
        except SkipBatchItemException as exc:
            LOG.info(str(exc))
            LOG.notice('=====> Jumping to next batch item.')
            continue
        except AbortException as exc:
            sys.exit(int(exc))
        except KeyboardInterrupt:
            print()
            LOG.warning('Aborted by user. Exiting.')
            sys.exit(130)


def create_auto_files(script: dict, batch_items: list, args: Namespace, tempdir: str):
    cmd_class = get_command_class()
    LOG.info(f'Using temporary directory to save object files: {tempdir}')
    digits = len(str(len(batch_items)))
    for i, row in enumerate(batch_items, start=1):
        script_copy = deepcopy(script)
        update_script_from_row(row=row, script=script_copy, index=i)

        variables = collect_vars(script_copy)

        auto = Automatix(
            script=script_copy,
            variables=variables,
            config=CONFIG,
            cmd_class=cmd_class,
            script_fields=SCRIPT_FIELDS,
            cmd_args=args,
            batch_index=1,
        )
        auto.set_command_count()
        id = str(i).rjust(digits, '0')
        with open(f'{tempdir}/auto{id}', 'wb') as f:
            pickle.dump(obj=auto, file=f)


def run_parallel_screens(script: dict, batch_items: list, args: Namespace):
    LOG.info('Preparing automatix objects for parallel processing')

    with TemporaryDirectory() as tempdir:
        create_auto_files(script=script, batch_items=batch_items, args=args, tempdir=tempdir)

        time_id = round(time())
        os.mkfifo(f'{tempdir}/{time_id}_finished')

        logfile_dir = f'{CONFIG.get("logfile_dir")}/{time_id}'
        os.makedirs(logfile_dir)
        LOG.info(f'Created directory for logfiles at {logfile_dir}')

        subprocess.run(
            f'screen -d -m -S {time_id}_overview'
            f' automatix-manager {tempdir} {time_id} {"--debug" if args.debug else ""}',
            shell=True,
        )

        LOG.info(f'Overview / manager screen started at "{time_id}_overview".')

        LOG.info('Start loop with information to switch between running screens.\n')
        screen_switch_loop(tempdir=tempdir, time_id=time_id)

        with open(f'{tempdir}/{time_id}_finished') as fifo:
            print()
            LOG.info('Wait for overview to finish')
            for _ in fifo:
                LOG.info('Automatix finished parallel processing')
    LOG.info('Temporary directory cleaned up')
    LOG.info(f'All logfiles are available at {logfile_dir}')


def main():
    check_for_original_automatix()

    if os.getenv('AUTOMATIX_SHELL'):
        print('You are running Automatix from an interactive shell of an already running Automatix!')
        answer = input('Do you really want to proceed? Then type "yes" and ENTER.\n')
        if answer != 'yes':
            sys.exit(0)

    starttime = time()
    args = arguments()
    setup(args=args)

    script, batch_items = get_script_and_batch_items(args=args)

    if args.vars_file and args.parallel:
        run_parallel_screens(script=script, batch_items=batch_items, args=args)
        sys.exit(0)

    try:
        if PROGRESS_BAR:
            progress_bar.setup_scroll_area()

        run_batch_items(script=script, batch_items=batch_items, args=args)
    finally:
        if PROGRESS_BAR:
            progress_bar.destroy_scroll_area()

    if 'AUTOMATIX_TIME' in os.environ:
        LOG.info(f'The Automatix script took {round(time() - starttime)}s!')



import argparse
import os
import pickle
import select
import subprocess
import sys
from dataclasses import dataclass, field
from os import listdir, unlink
from os.path import isfile
from time import sleep

from .automatix import Automatix
from .colors import yellow, green, red, cyan
from .command import AbortException
from .config import LOG, init_logger, CONFIG, PROGRESS_BAR, progress_bar

LINE_UP = '\033[1A'
LINE_CLEAR = '\x1b[2K'


@dataclass
class Autos:
    status_file: str
    time_id: int
    count: int

    max_parallel: int = 2
    waiting: set = field(default_factory=set)
    running: set = field(default_factory=set)
    user_input: set = field(default_factory=set)
    finished: set = field(default_factory=set)


class FileWithLock:
    def __init__(self, file_path: str, method: str):
        self.file_path = file_path
        self.method = method
        self.file_obj = None

    def __enter__(self):
        get_lock(self.file_path)
        self.file_obj = open(self.file_path, self.method)
        return self.file_obj

    def __exit__(self, type, value, traceback):
        self.file_obj.close()
        release_lock(self.file_path)


def get_lock(file_path: str):
    while True:
        try:
            os.mkdir(f'{file_path}.lock')
        except FileExistsError:
            sleep(1)
            continue
        break


def release_lock(file_path: str):
    os.rmdir(f'{file_path}.lock')


def get_files(tempdir: str) -> set:
    return {f for f in listdir(tempdir) if isfile(f'{tempdir}/{f}') and f.startswith('auto')}


def print_status(autos: Autos):
    tmpl = 'waiting: {w}, running: {r}, user input required: {u}, finished: {f}'
    print(tmpl.format(
        w=yellow(len(autos.waiting)),
        r=cyan(len(autos.running)),
        u=red(len(autos.user_input)),
        f=green(f'{len(autos.finished)}/{autos.count}'),
    ))


def print_status_verbose(autos: Autos):
    print(f'------------------ Screens (max. {autos.max_parallel} running) ------------------')
    print_status(autos=autos)
    print('--------------------------------------------------------------')
    print(f'waiting:             {sorted(autos.waiting)}')
    print(f'running:             {sorted(autos.running)}')
    print(f'user input required: {red(sorted(autos.user_input))}')
    print(f'finished:            {sorted(autos.finished)}')


def check_for_status_change(autos: Autos, status_file: str):
    with FileWithLock(status_file, 'r+') as sf:
        for line in sf:
            if not line:
                continue
            LOG.debug(f'Line: {line}')
            auto_file, status = line.strip().split(':')
            LOG.debug(f'Got {auto_file}:{status}')
            match status:
                case 'max_parallel':
                    # In this case we misuse the "auto_file" part as number
                    # for how many parallel screens are allowed.
                    autos.max_parallel = int(auto_file)
                    LOG.info(f'Now process max {auto_file} screens parallel')
                case 'user_input_remove':
                    autos.user_input.remove(auto_file)
                case 'user_input_add':
                    autos.user_input.add(auto_file)
                    LOG.info(f'{auto_file} is waiting for user input')
                case 'finished':
                    autos.running.remove(auto_file)
                    autos.finished.add(auto_file)
                    LOG.info(f'{auto_file} finished')
                case _:
                    LOG.warning(f'[{auto_file}] Unrecognized status "{status}"\n')
        sf.truncate(0)


def run_manage_loop(tempdir: str, time_id: int):
    status_file = f'{tempdir}/{time_id}_overview'
    auto_files = get_files(tempdir)
    autos = Autos(status_file=status_file, time_id=time_id, count=len(auto_files), waiting=auto_files)
    logfile_dir = f'{CONFIG.get("logfile_dir")}/{time_id}'

    LOG.info(f'Found {autos.count} files to process. Screens name are like "{time_id}_autoX"')
    LOG.info('To switch to screens detach from this screen via "<ctrl>+a d".')

    open(status_file, 'a').close()
    try:
        while len(autos.finished) < autos.count:
            if len(autos.running) < autos.max_parallel and autos.waiting:
                auto_file = autos.waiting.pop()
                autos.running.add(auto_file)
                LOG.info(f'Starting new screen at {time_id}_{auto_file}')
                subprocess.run(
                    f'screen -d -m -S {time_id}_{auto_file}'
                    f' -L -Logfile {logfile_dir}/{auto_file}.log'
                    f' automatix-from-file {tempdir} {time_id} {auto_file}',
                    # for debugging replace line above with:
                    # f' bash -c "automatix-from-file {tempdir} {time_id} {auto_file} || bash"',
                    shell=True,
                )

            check_for_status_change(autos=autos, status_file=status_file)

            print_status(autos=autos)

            # Write status to file for screen switch loop
            with FileWithLock(f'{tempdir}/autos', 'wb') as f:
                pickle.dump(obj=autos, file=f)

            sleep(1)

        LOG.info(f'All parallel screen reported finished ({len(autos.finished)}/{autos.count}).')
    except Exception as exc:
        LOG.exception(exc)
        sleep(60)  # For debugging
    finally:
        with open(f'{tempdir}/{time_id}_finished', 'w') as fifo:
            fifo.write('finished')


def run_manager():
    parser = argparse.ArgumentParser(
        description='Automatix manager for parallel processing (called by the automatix main programm)',
        epilog='Explanations and README at https://github.com/vanadinit/automatix',
    )
    parser.add_argument('tempdir')
    parser.add_argument('time_id')
    parser.add_argument(
        '--debug', '-d',
        action='store_true',
        help='activate debug log level',
    )
    args = parser.parse_args()

    init_logger(name=CONFIG['logger'], debug=args.debug)
    run_manage_loop(tempdir=args.tempdir, time_id=int(args.time_id))


def run_auto(tempdir: str, time_id: int, auto_file: str):
    auto_path = f'{tempdir}/{auto_file}'

    def send_status(status: str):
        with FileWithLock(f'{tempdir}/{time_id}_overview', 'a') as sf:
            sf.write(f'{auto_file}:{status}\n')

    with open(auto_path, 'rb') as f:
        auto: Automatix = pickle.load(file=f)
    auto.env.attach_logger()
    auto.env.reinit_logger()
    auto.env.send_status = send_status

    try:
        auto.run()
    except AbortException as exc:
        sys.exit(int(exc))
    except KeyboardInterrupt:
        print()
        LOG.warning('Aborted by user. Exiting.')
        sys.exit(130)
    finally:
        send_status('finished')
        unlink(auto_path)


def run_auto_from_file():
    parser = argparse.ArgumentParser(
        description='Automatix from file for parallel processing (called by the automatix-manager)',
        epilog='Explanations and README at https://github.com/vanadinit/automatix',
    )
    parser.add_argument('tempdir')
    parser.add_argument('time_id')
    parser.add_argument('auto_file')
    args = parser.parse_args()

    try:
        if PROGRESS_BAR:
            progress_bar.setup_scroll_area()

        run_auto(tempdir=args.tempdir, time_id=args.time_id, auto_file=args.auto_file)
    finally:
        if PROGRESS_BAR:
            progress_bar.destroy_scroll_area()


def ask_for_options(autos: Autos) -> str | None:
    print()
    LOG.notice('Please notice: To come back to this selection press "<ctrl>+a d" in a screen session!')
    LOG.info('Following options are available:')
    LOG.info(
        'o: overview / manager loop,'
        ' n: next user input required,'
        ' X (number): switch to autoX,'
        f' mX: set max parallel screens to X (actual {autos.max_parallel})\n'
    )
    i, _, _ = select.select([sys.stdin], [], [], 1)
    if i:
        answer = sys.stdin.readline().strip()
    else:
        return

    if answer == '':
        return

    if answer == 'o':
        return f'{autos.time_id}_overview'

    if answer == 'n' and autos.user_input:
        return f'{autos.time_id}_{next(iter(autos.user_input))}'

    if answer.startswith('m'):
        try:
            max_parallel = int(answer[1:])
            with FileWithLock(autos.status_file, 'a') as sf:
                sf.write(f'{max_parallel}:max_parallel\n')
            return
        except ValueError:
            LOG.warning(f'Invalid answer: {answer}')
            sleep(1)
            return

    try:
        number = int(answer)
        return f'{autos.time_id}_auto{str(number).rjust(len(str(autos.count)), "0")}'
    except ValueError:
        LOG.warning(f'Invalid answer: {answer}')
        sleep(1)
        return


def screen_switch_loop(tempdir: str, time_id: int):
    try:
        while not isfile(f'{tempdir}/autos'):
            sleep(1)
        while True:
            with FileWithLock(f'{tempdir}/autos', 'rb') as f:
                autos = pickle.load(file=f)
            print_status_verbose(autos=autos)

            if len(autos.running) + len(autos.waiting) + len(autos.user_input) == 0:
                break

            if screen_id := ask_for_options(autos=autos):
                subprocess.run(f'screen -r {screen_id}', shell=True)
                sleep(1)
            else:
                for _ in range(12):
                    print(LINE_UP, end=LINE_CLEAR)
    except (KeyboardInterrupt, Exception) as exc:
        print('\n' * 8)
        LOG.exception(exc)
        print()
        LOG.info('An exception occurred! Please decide what to do:')
        match input(
            'Press "r" and Enter to reraise.'
            ' This will cause this programm to terminate.'
            ' Check "screen -list" afterwards for still running screens.\n'
            ' Press something else to restart the loop, where you can switch to the different screens.\n'
        ):
            case 'r':
                raise
            case _:
                screen_switch_loop(tempdir=tempdir, time_id=time_id)



import logging

from sys import stderr, stdout

NOTICE = 25

C_RED = '\033[91m'
C_BLUE = '\033[94m'
C_END = '\033[0m'


class LevelFilter(logging.Filter):

    def __init__(self, min_level: int = logging.DEBUG, max_level: int = logging.WARNING) -> None:
        super().__init__()
        self._min_level = min_level
        self._max_level = max_level

    def filter(self, record: logging.LogRecord) -> bool:
        return self._min_level <= record.levelno <= self._max_level


class ErrorFormatter(logging.Formatter):

    def format(self, record: logging.LogRecord) -> str:
        if record.levelno < logging.WARNING:
            return logging.Formatter.format(self, record)

        original_format = self._style._fmt
        self._style._fmt = C_RED + '%(name)s: ' + original_format + C_END

        formatted_record = logging.Formatter.format(self, record)
        self._style._fmt = original_format
        return formatted_record


class ConsoleFormatter(logging.Formatter):

    def format(self, record: logging.LogRecord) -> str:
        if record.levelno != NOTICE:
            return logging.Formatter.format(self, record)

        original_format = self._style._fmt
        self._style._fmt = C_BLUE + original_format + C_END

        formatted_record = logging.Formatter.format(self, record)
        self._style._fmt = original_format
        return formatted_record


def init_logger(name: str = 'automatix', debug: bool = False):
    log = logging.getLogger(name=name)
    if log.hasHandlers():
        return

    log.setLevel(logging.DEBUG)  # accept everything and let handlers decide

    for handler in _setup_handlers(debug=debug):
        log.addHandler(handler)


def _setup_handlers(debug: bool = False) -> [logging.Handler]:
    handlers = []
    console_handler = logging.StreamHandler(stream=stdout)
    console_handler.setLevel(logging.DEBUG)
    error_handler = logging.StreamHandler(stream=stderr)
    error_handler.setLevel(logging.WARNING)

    handlers.append(error_handler)
    handlers.append(console_handler)

    if debug:
        min_level = logging.DEBUG
        base_format = '%(asctime)s [ %(levelname)s ] %(message)s (%(filename)s:%(lineno)d)'
    else:
        min_level = logging.INFO
        base_format = '%(message)s'
    console_handler.addFilter(LevelFilter(min_level=min_level, max_level=NOTICE))
    console_handler.setFormatter(ConsoleFormatter(fmt=base_format))
    error_handler.setFormatter(ErrorFormatter(fmt=base_format))

    return handlers


def _patch_add_notice_level_to_logging() -> None:
    logging.addLevelName(NOTICE, 'NOTICE')
    logging.Logger.notice = _notice


def _notice(self, msg, *args, **kwargs):
    if self.isEnabledFor(NOTICE):
        self._log(NOTICE, msg, args, **kwargs)


_patch_add_notice_level_to_logging()



import argparse
import logging
import os
import re
import sys
from collections import OrderedDict
from importlib import metadata, import_module
from time import sleep

import yaml

yaml.warnings({'YAMLLoadWarning': False})


def read_yaml(yamlfile: str) -> dict:
    with open(yamlfile) as file:
        return yaml.load(file.read(), Loader=yaml.SafeLoader)


try:
    from argcomplete import autocomplete
    from .bash_completion import ScriptFileCompleter

    bash_completion = True
except ImportError:
    bash_completion = False

VERSION = metadata.version('automatix_cmd')

DEPRECATED_SYNTAX = {
    # 0: REGEX pattern
    # 1: replacement, formatted with group = re.Match.groups(), e.g. 'something {group[0]} foo'
    # 2: special flags (p: python, b: Bundlewrap, s: replace '{s}' with pipe separated system names)
    (r'({s})_node(?!\w)', 'NODES.{group[0]}', 'bps'),  # Removed in 2.0.0
    (r'{\s*system_(\w*)\s*}', '{{SYSTEMS.{group[0]}}}', ''),  # Removed in 2.0.0
    (r'{\s*const_(\w*)\s*}', '{{CONST.{group[0]}}}', ''),  # Removed in 2.0.0
    (r'(?<!\w)global\s+(\w*)', 'PERSISTENT_VARS[\'{group[0]}\'] = {group[0]}', 'p'),
}

SCRIPT_FIELDS = OrderedDict()
SCRIPT_FIELDS['systems'] = 'Systems'
SCRIPT_FIELDS['vars'] = 'Variables'

CONFIG = {
    'script_dir': '~/automatix-config',
    'constants': {},
    'encoding': 'utf-8',
    'import_path': '.',
    'ssh_cmd': 'ssh {hostname} sudo ',
    'remote_tmp_dir': 'automatix_tmp',
    'logger': 'automatix',
    'logfile_dir': 'automatix_logs',
    'bundlewrap': False,
    'teamvault': False,
}

configfile = os.path.expanduser(os.path.expandvars(os.getenv('AUTOMATIX_CONFIG', '~/.automatix.cfg.yaml')))
if os.path.isfile(configfile):
    CONFIG.update(read_yaml(configfile))
    CONFIG['config_file'] = configfile

for key, value in CONFIG.items():
    if not isinstance(value, str):
        continue
    if os.getenv(f'AUTOMATIX_{key.upper()}'):
        CONFIG[key] = os.getenv(f'AUTOMATIX_{key.upper()}')

if CONFIG.get('logging_lib'):
    log_lib = import_module(CONFIG.get('logging_lib'))
    init_logger = log_lib.init_logger  # noqa F401
else:
    from .logger import init_logger  # noqa F401

LOG = logging.getLogger(CONFIG['logger'])

SCRIPT_PATH = os.path.expanduser(os.path.expandvars(CONFIG['script_dir']))

if CONFIG['teamvault']:
    import bwtv

    SCRIPT_FIELDS['secrets'] = 'Secrets'

    class UnknownSecretTypeException(Exception):
        pass

progress_bar = None
PROGRESS_BAR = False
if (CONFIG.get('progress_bar', False) or os.getenv('AUTOMATIX_PROGRESS_BAR', False)) and \
        os.getenv('AUTOMATIX_PROGRESS_BAR', False) not in ['False', 'false']:
    try:
        import python_progress_bar as progress_bar  # noqa F401

        PROGRESS_BAR = True
    except ImportError:
        pass


def arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description='Automation wrapper for bash and python commands.',
        epilog='Explanations and README at https://github.com/vanadinit/automatix',
    )
    scriptfile_action = parser.add_argument(
        'scriptfile',
        help='Path to scriptfile (yaml), use " -- " if needed to delimit this from argument fields',
    )
    if bash_completion:
        scriptfile_action.completer = ScriptFileCompleter(script_path=SCRIPT_PATH)

    for field in SCRIPT_FIELDS.keys():
        parser.add_argument(
            f'--{field}',
            nargs='*',
            help=f'Use this to set {field} without adding them to the script or to overwrite them. '
                 f'You can specify multiple {field} like: --{field} v1=string1 v2=string2 v3=string3',
        )
    parser.add_argument(
        '--vars-file',
        help='Path to a CSV file containing variables for batch processing',
    )
    parser.add_argument(
        '--parallel',
        action='store_true',
        help='Run CSV file entries parallel in screen sessions; only valid with --vars-file. '
             'GNU screen has to be installed. See EXTRAS section in README.',
    )
    parser.add_argument(
        '--print-overview', '-p',
        action='store_true',
        help='Just print command pipeline overview with indices then exit without executing the commandline. '
             'Note that the *always pipeline* will be executed beforehand.',
    )
    parser.add_argument(
        '--jump-to', '-j',
        help='Jump to step instead of starting at the beginning',
        type=int,
        default=0,
    )
    parser.add_argument(
        '--steps', '-s',
        help='Only execute these steps (comma-separated indices) or exclude steps (prepend "e")',
    )
    parser.add_argument(
        '--interactive', '-i',
        action='store_true',
        help='Confirm actions before executing',
    )
    parser.add_argument(
        '--force', '-f',
        action='store_true',
        help='try always to proceed (except manual steps), even if errors occur (no retries)'
    )
    parser.add_argument(
        '--debug', '-d',
        action='store_true',
        help='activate debug log level',
    )
    if bash_completion:
        autocomplete(parser)
    return parser.parse_args()


def _overwrite(script: dict, key: str, data: list[str]):
    script.setdefault(key, {})
    for item in data:
        k, v = item.split('=')
        script[key][k] = v


def _tupelize(string) -> tuple:
    return tuple([int(i) for i in string.split('.')])


def get_script(args: argparse.Namespace) -> dict:
    s_file = args.scriptfile
    if not os.path.isfile(args.scriptfile):
        s_file = f'{SCRIPT_PATH}/{args.scriptfile}'

    try:
        script = read_yaml(s_file)
        validate_script(script)
    except Exception:
        LOG.exception('Script validation failed! Please fix syntax before retrying!')
        if input('To reload and proceed after fixing type "R" and press Enter.\a') == 'R':
            return get_script(args=args)
        sys.exit(1)

    for field in SCRIPT_FIELDS.keys():
        if vars(args).get(field):
            _overwrite(script=script, key=field, data=vars(args)[field])

    return script


def check_deprecated_syntax(ckey: str, entry: str, script: dict, prefix: str) -> int:
    warn = 0
    if isinstance(entry, dict):
        warn += 1
        LOG.warning(f'{prefix} Command is not a string! Please use quotes!')
        entry = f'{{{next(iter(entry))}}}'

    for pattern, replacement, flags in DEPRECATED_SYNTAX:
        if 'b' in flags and not CONFIG['bundlewrap']:
            continue
        if 'p' in flags and 'python' not in ckey:
            continue
        if 's' in flags:
            match = re.search(pattern.format(s='|'.join(script.get('systems', {}).keys())), entry)
        else:
            match = re.search(pattern, entry)
        if match:
            warn += 1
            LOG.warning('{prefix} Using "{match}" is deprecated. Use "{repl}" instead.'.format(
                prefix=prefix,
                match=match.group(0),
                repl=replacement.format(group=match.groups())
            ))
    return warn


def validate_script(script: dict):
    script_required_version = script.get('require_version', '0.0.0')
    if _tupelize(VERSION) < _tupelize(script_required_version):
        LOG.error(f'The script requires minimum version {script_required_version}. We have {VERSION}.')
        sys.exit(1)
    warn = 0
    for pipeline in ['always', 'pipeline', 'cleanup']:
        for index, command in enumerate(script.get(pipeline, [])):
            for ckey, entry in command.items():
                warn += check_deprecated_syntax(ckey=ckey, entry=entry, script=script, prefix=f'[{pipeline}:{index}]')
                break  # there always should be only one entry
    if warn:
        # To give people a chance to see warnings before the following output happens.
        sleep(5)


def collect_vars(script: dict) -> dict:
    var_dict = script.get('vars', {})
    if var_dict is None:
        LOG.warning('Vars section defined, but empty!\nThis is illegal, either remove the section or add variables.')
        var_dict = {}
    script['vars'] = var_dict  # just for the case it was empty
    if CONFIG['teamvault']:
        for key, secret in script.get('secrets', {}).items():
            sid, field = secret.split('_')
            if field == 'password':
                var_dict[key] = bwtv.password(sid)
            elif field == 'username':
                var_dict[key] = bwtv.username(sid)
            elif field == 'file':
                var_dict[key] = bwtv.file(sid)
            else:
                raise UnknownSecretTypeException(field)
    return var_dict


def update_script_from_row(row: dict, script: dict, index: int):
    if not row:
        return
    try:
        script['name'] += f" ({index} {row.pop('label')})"
    except KeyError:
        script['name'] += f" ({index})"
    for key, value in row.items():
        assert len(key.split(':')) == 2, \
            'First row in CSV must contain "label" or the field name and key seperated by colons' \
            ' like "label,systems:mysystem,vars:myvar".'
        key_type, key_name = key.split(':')
        assert key_type in SCRIPT_FIELDS.keys(), \
            f'First row in CSV: Field name is \'{key_type}\', but has to be one of {list(SCRIPT_FIELDS.keys())}.'
        script[key_type][key_name] = value



from automatix.command import Command
from tests.test_environment import testauto, environment


def test__automatix__command_list():
    assert testauto.command_list('main') == testauto.command_list('pipeline')
    cmd = Command(
        cmd={'local': "echo 'Print this always :-)'"},
        index=0,
        pipeline='always',
        env=environment,
        position=0,
    )
    first_cleanup = testauto.command_list('always')[0]
    assert isinstance(first_cleanup, Command)
    assert vars(first_cleanup) == vars(cmd)


def test__automatix__get_command_position():
    assert testauto.get_command_position(index=4, pipeline='always') == 4
    assert testauto.get_command_position(index=3, pipeline='always') != 4
    assert testauto.get_command_position(index=3, pipeline='main') == 5  # 2 commands in always pipeline
    assert testauto.get_command_position(index=3, pipeline='pipeline') == 5  # 2 commands in always pipeline
    assert testauto.get_command_position(index=2, pipeline='cleanup') == 22  # 20 commands in always + main pipeline


def test__automatix__set_command_count():
    testauto.env.command_count = None
    testauto.set_command_count()
    assert testauto.env.command_count == 21



from argparse import Namespace
from logging import getLogger

from .config import init_logger


class PipelineEnvironment:
    def __init__(
            self,
            name: str,
            config: dict,
            systems: dict,
            variables: dict,
            imports: list,
            batch_mode: bool,
            batch_items_count: int,
            batch_index: int,
            cmd_args: Namespace,
    ):
        self.name = name
        self.config = config
        self.systems = systems
        self.vars = variables
        self.imports = imports
        self.batch_mode = batch_mode
        self.batch_items_count = batch_items_count
        self.batch_index = batch_index
        self.cmd_args = cmd_args
        self.LOG = None

        # This will be set at runtime
        self.command_count = None

    def attach_logger(self):
        self.LOG = getLogger(self.config['logger'])

    def reinit_logger(self):
        self.LOG.handlers.clear()
        init_logger(name=self.LOG.name, debug=self.cmd_args.debug)

    def send_status(self, status: str):
        return



import os
import re
import subprocess
from shlex import quote
from time import time
from typing import Tuple

from .config import PROGRESS_BAR, progress_bar
from .environment import PipelineEnvironment


class PersistentDict(dict):
    def __getattr__(self, key: str):
        return self[key]

    def __hasattr__(self, key: str):
        return key in self

    def __setattr__(self, key: str, value):
        self[key] = value


PERSISTENT_VARS = PVARS = PersistentDict()

SHELL_EXECUTABLE = '/bin/bash'

# Leading red " Automatix \w > " to indicate that this shell is inside a running Automatix execution
AUTOMATIX_PROMPT = r'\[\033[0;31m\] Automatix \[\033[0m\]\\w > '

POSSIBLE_ANSWERS = {
    'p': 'proceed (default)',
    'T': f'start interactive terminal shell ({SHELL_EXECUTABLE} -i) and return back here on exit',
    'v': 'show and change variables',
    'r': 'retry',
    'R': 'reload from file and retry command (same index)',
    'R±X': 'same as R, but change index by X (integer)',
    's': 'skip',
    'a': 'abort',
    'c': 'abort & continue to next (CSV processing)',
    # 't' -> reserved for handling still running remote processes
    # 'k' -> reserved for handling still running remote processes
    # 'i' -> reserved for handling still running remote processes
}

KEYBOARD_INTERRUPT_MESSAGE = 'Abort command by user key stroke. Exit code is set to 130.'


class Command:
    def __init__(self, cmd: dict, index: int, pipeline: str, env: PipelineEnvironment, position: int):
        self.cmd = cmd
        self.index = index
        self.env = env
        self.pipeline = pipeline
        self.position = position

        for key, value in cmd.items():
            self.orig_key = key
            self.condition_var, self.assignment_var, self.key = parse_key(key=key)
            if isinstance(value, dict):
                # We need this workaround because the yaml lib returns a dictionary instead of a string,
                # if there is nothing but a variable in the command. Alternative is to use quotes in the script yaml.
                self.value = f'{{{next(iter(value))}}}'
            else:
                self.value = value
            break  # There should be only one entry in pipeline_cmd

    @property
    def progress_portion(self) -> int:
        own_position = self.env.command_count * (self.env.batch_index - 1) + self.position
        overall_command_count = self.env.batch_items_count * self.env.command_count
        return round(own_position / overall_command_count * 100, 1)

    def get_type(self):
        if self.key == 'local':
            return 'local'
        if self.key == 'manual':
            return 'manual'
        if self.key == 'python':
            return 'python'
        if re.search('remote@', self.key):
            return 'remote'
        raise UnknownCommandException(f'Command type {self.key} is not known.')

    def get_system(self):
        if self.get_type() == 'remote':
            return self.env.systems[re.search(r'remote@(.*)', self.key).group(1)]
        return 'localhost'

    def get_resolved_value(self):
        variables = self.env.vars.copy()
        for key, value in variables.items():
            if not isinstance(value, str):
                continue
            file_match = re.match(r'FILE_(.*)', value)
            if file_match:
                with open(os.path.expanduser(file_match.group(1))) as file:
                    variables[key] = file.read().strip()

        variables['CONST'] = ConstantsWrapper(self.env.config['constants'])
        variables['SYSTEMS'] = SystemsWrapper(self.env.systems)
        return self.value.format(**variables)

    def print_command(self):
        print()
        self.env.LOG.notice(f'({self.index}) [{self.orig_key}]: {self.get_resolved_value()}')

    def show_and_change_variables(self):
        print()
        self.env.LOG.info('Variables:')
        for key, value in self.env.vars.items():
            self.env.LOG.info(f" {key}: {value}")
        print()
        self.env.LOG.info('To change/set variable write variable + "=" followed by value.')
        self.env.LOG.info('Example: var1=xyz')
        self.env.LOG.info('Notice: You can only change 1 variable at a time. Repeat if necessary.')
        self.env.LOG.info('To not change anything just press "ENTER".')
        answer = input('\n')
        try:
            key, value = answer.split('=', maxsplit=1)
            self.env.vars[key.strip()] = value.strip()
            self.env.LOG.info(f'Variable {key.strip()} = {value.strip()}')
        except ValueError:
            if answer:
                self.env.LOG.error('Input could not be parsed.')
        self.print_command()
        print()

    def execute(self, interactive: bool = False, force: bool = False):
        try:
            self._execute(interactive=interactive, force=force)
        except (KeyError, UnknownCommandException):
            self.env.LOG.exception('Syntax or value error!')
            self.env.LOG.error('Syntax or value error! Please fix your script and reload/restart.')
            self._ask_user(question='[SE] What should I do?', allowed_options=['R', 'T', 'v', 's', 'a'])
        if PROGRESS_BAR:
            progress_bar.draw_progress_bar(self.progress_portion)

    def _execute(self, interactive: bool = False, force: bool = False):
        self.print_command()

        if not self._check_condition():
            self.env.LOG.info('Skip command, because the condition is not met.')
            return

        if self.get_type() == 'manual' or interactive:
            self.env.LOG.debug('Ask for user interaction.')

            answer = self._ask_user(question='[MS] Proceed?', allowed_options=['p', 'T', 'v', 's', 'R', 'a'])
            # answers 'a', 'c' and 'R' are handled by _ask_user, 'p' means just pass
            if answer == 's' or self.get_type() == 'manual':
                return

        steptime = time()

        return_code = self._execute_action()

        if 'AUTOMATIX_TIME' in os.environ:
            print()
            self.env.LOG.info(f'(command execution time: {round(time() - steptime)}s)')

        if return_code != 0:
            self.env.LOG.error(
                f'>> {self.env.name} << Command ({self.pipeline}:{self.index}) failed with return code {return_code}.')
            if force:
                return

            err_answer = self._ask_user(
                question='[CF] What should I do?',
                allowed_options=['p', 'T', 'v', 'r', 'R', 'a'],
            )
            # answers 'T', 'v', 'a', 'c' and 'R' are handled by _ask_user, 'p' means just pass
            if err_answer == 'r':
                return self._execute(interactive)

    def _check_condition(self) -> bool:
        if self.condition_var is None:
            return True

        if self.condition_var.endswith('!'):
            cond_var = self.condition_var[:-1]
            invert = True
        else:
            cond_var = self.condition_var
            invert = False

        if cond_var.startswith('PVARS.'):
            condition = PERSISTENT_VARS[cond_var[6:]]
        else:
            condition = self.env.vars.get(cond_var, False)

        return bool(condition) != invert

    def _execute_action(self) -> int:
        self.env.LOG.info('>')
        if self.get_type() == 'local':
            return self._local_action()
        if self.get_type() == 'python':
            return self._python_action()
        if self.get_type() == 'remote':
            return self._remote_action()

    def _ask_user(self, question: str, allowed_options: list) -> str:
        """
        User-Interactive handling of different scenarios.
        Questions should be prefixed with:
        [CF] Command Failed
        [PF] Partial command Failed (BW groups)
        [MS] Manual Step
        [RR] Remote process still Running
        [SE] Syntax Error

        :param question:
        :param allowed_options: character list of POSSIBLE_ANSWERS
        :return: answer character
        """
        if self.env.batch_mode:
            allowed_options.append('c')

        if 'R' in allowed_options:
            allowed_options.insert(allowed_options.index('R') + 1, 'R±X')

        options = '\n'.join([f' {k}: {POSSIBLE_ANSWERS[k]}' for k in allowed_options])

        return self._ask_user_with_options(
            question=f'{question}\n{options}\nYour answer: \a',
            allowed_options=allowed_options,
        )

    def _ask_user_with_options(self, question: str, allowed_options: list) -> str:
        if PROGRESS_BAR:
            progress_bar.block_progress_bar(self.progress_portion)
        self.env.send_status('user_input_add')
        answer = input(question)
        self.env.send_status('user_input_remove')

        if answer == '':  # default
            answer = 'p'

        if answer.startswith('R') and 'R' in allowed_options and len(answer) > 1:
            try:
                raise ReloadFromFile(index=self.index + int(answer[1:]))
            except ValueError:
                pass

        if answer not in allowed_options:
            self.env.LOG.warning('Invalid input. Try again.')
            return self._ask_user_with_options(question=question, allowed_options=allowed_options)

        match answer:
            case 'T':
                print()
                self.env.LOG.notice('Starting interactive terminal shell')
                self._run_local_command(
                    f'AUTOMATIX_SHELL=True'
                    f' {SHELL_EXECUTABLE}'
                    f' --rcfile <(cat ~/.bashrc ; echo "PS1=\\"{AUTOMATIX_PROMPT}\\"")'
                    f' -i'
                )
                return self._ask_user_with_options(question=question, allowed_options=allowed_options)
            case 'v':
                self.show_and_change_variables()
                return self._ask_user_with_options(question=question, allowed_options=allowed_options)
            case 'a':
                raise AbortException(1)
            case 'R':
                raise ReloadFromFile(index=self.index)
            case 'c':
                raise SkipBatchItemException()
            case _:
                return answer

    def _generate_python_vars(self):
        # For BWCommand this method is overridden
        return {'vars': self.env.vars}

    def _python_action(self) -> int:
        cmd = self.get_resolved_value()
        locale_vars = self._generate_python_vars()
        locale_vars.update(PERSISTENT_VARS)
        locale_vars.update({
            'AbortException': AbortException,
            'SkipBatchItemException': SkipBatchItemException,
        })

        self.env.LOG.debug(f'locals:\n {locale_vars}')
        try:
            self.env.LOG.debug(f'Run python command: {cmd}')
            if self.assignment_var:
                exec(f'vars["{self.assignment_var}"] = {cmd}', globals(), locale_vars)
                self.env.LOG.info(f'Variable {self.assignment_var} = {repr(self.env.vars[self.assignment_var])}')
            else:
                exec(cmd, globals(), locale_vars)
            return 0
        except (AbortException, SkipBatchItemException):
            raise
        except KeyboardInterrupt:
            self.env.LOG.info(KEYBOARD_INTERRUPT_MESSAGE)
            return 130
        except Exception as exc:
            if isinstance(exc, NameError) and not self.env.config.get('bundlewrap') and str(exc) in [
                'name \'NODES\' is not defined',
                'name \'AUTOMATIX_BW_REPO\' is not defined',
            ]:
                self.env.LOG.exception(
                    'Seems you are trying to use bundlewrap functions without having bundlewrap support enabled.'
                    ' Please check your configuration.')
                return 1
            if isinstance(exc.__context__, ReloadFromFile):
                exc.__suppress_context__ = True

            self.env.LOG.exception('Unknown error occured:')
            return 1

    def _local_action(self) -> int:
        cmd = self._build_command(path=self.env.config['import_path'])
        try:
            return self._run_local_command(cmd=cmd)
        except KeyboardInterrupt:
            self.env.LOG.info(KEYBOARD_INTERRUPT_MESSAGE)
            return 130

    def _build_command(self, path: str) -> str:
        if not self.env.imports:
            return self.get_resolved_value()
        return f'. {path}/' + f'; . {path}/'.join(self.env.imports) + '; ' + self.get_resolved_value()

    def _run_local_command(self, cmd: str) -> int:
        self.env.LOG.debug(f'Executing: {cmd}')
        if self.assignment_var:
            proc = subprocess.run(cmd, shell=True, executable=SHELL_EXECUTABLE, stdout=subprocess.PIPE)
            output = proc.stdout.decode(self.env.config["encoding"])
            self.env.vars[self.assignment_var] = assigned_value = output.rstrip('\r\n')
            hint = ' (trailing newline removed)' if (output.endswith('\n') or output.endswith('\r')) else ''
            self.env.LOG.info(f'Variable {self.assignment_var} = "{assigned_value}"{hint}')
        else:
            proc = subprocess.run(cmd, shell=True, executable=SHELL_EXECUTABLE)
        return proc.returncode

    def _remote_action(self) -> int:
        # For BWCommand this method is overridden
        return self._remote_action_on_hostname(hostname=self.get_system().replace('hostname!', ''))

    def _remote_action_on_hostname(self, hostname: str) -> int:
        try:
            exitcode = self._run_local_command(cmd=self._get_remote_command(hostname=hostname))
        except KeyboardInterrupt:
            self.env.LOG.info(KEYBOARD_INTERRUPT_MESSAGE)
            exitcode = 130
            self._remote_handle_keyboard_interrupt(hostname=hostname)

        if self.env.imports:
            self._remote_cleanup_imports(hostname=hostname)

        return exitcode

    def _get_remote_command(self, hostname: str) -> str:
        ssh_cmd = self.env.config["ssh_cmd"].format(hostname=hostname)
        remote_cmd = self.get_resolved_value()
        prefix = ''
        if self.env.imports:
            # How is this working?
            # - Create a tar archive with all imports
            # - Pipe it through SSH
            # - Create tmp dir on remote
            # - Extract tar archive there
            # - Source imports in _build_command
            prefix = f'tar -C {self.env.config["import_path"]} -cf - ' + ' '.join(self.env.imports) + ' | '
            remote_cmd = f'mkdir {self.env.config["remote_tmp_dir"]};' \
                         f' tar -C {self.env.config["remote_tmp_dir"]} -xf -;' \
                         f' {self._build_command(path=self.env.config["remote_tmp_dir"])}'

        return f'{prefix}{ssh_cmd}{quote("bash -c " + quote(remote_cmd))}'

    def _remote_handle_keyboard_interrupt(self, hostname: str) -> int:
        ssh_cmd = self.env.config["ssh_cmd"].format(hostname=hostname)

        try:
            ps_pids = self.get_remote_pids(hostname=hostname, cmd=self.get_resolved_value())
            while ps_pids:
                self.env.LOG.notice(
                    'Remote command seems still to be running! Found PIDs: {}'.format(','.join(ps_pids))
                )
                self.env.send_status('user_input_add')
                answer = input(
                    '[RR] What should I do? '
                    '(i: send SIGINT (default), t: send SIGTERM, k: send SIGKILL, p: do nothing and proceed) \n\a')
                self.env.send_status('user_input_remove')

                if answer == 'p':
                    break
                elif answer == 't':
                    signal = 'TERM'
                elif answer == 'k':
                    signal = 'KILL'
                else:
                    signal = 'INT'
                for pid in ps_pids:
                    kill_cmd = f'{ssh_cmd} kill -{signal} {pid}'
                    self.env.LOG.info(f'Kill {pid} on {hostname}')
                    subprocess.run(kill_cmd, shell=True)

                ps_pids = self.get_remote_pids(hostname=hostname, cmd=self.get_resolved_value())
        except subprocess.CalledProcessError:
            self.env.LOG.warning('Could not check for remaining remote processes.')

        self.env.LOG.info('Keystroke interrupt handled.\n')

    def get_remote_pids(self, hostname, cmd) -> []:
        ps_cmd = f"ps axu | grep {quote(cmd)} | grep -v 'grep' | awk '{{print $2}}'"
        cmd = f'ssh {hostname} {quote(ps_cmd)} 2>&1'
        pids = subprocess.check_output(
            cmd,
            shell=True,
            executable=SHELL_EXECUTABLE,
        ).decode(self.env.config["encoding"]).split()

        return pids

    def _remote_cleanup_imports(self, hostname: str):
        ssh_cmd = self.env.config["ssh_cmd"].format(hostname=hostname)
        cleanup_cmd = f'{ssh_cmd} rm -r {self.env.config["remote_tmp_dir"]}'
        self.env.LOG.debug(f'Executing: {cleanup_cmd}')
        proc = subprocess.run(cleanup_cmd, shell=True, executable=SHELL_EXECUTABLE)
        if proc.returncode != 0:
            self.env.LOG.warning(
                'Failed to remove {tmp_dir}, exitcode: {return_code}'.format(
                    tmp_dir=self.env.config["remote_tmp_dir"],
                    return_code=proc.returncode,
                )
            )


def parse_key(key) -> Tuple[str, ...]:
    """
    parses the key

    returns a list containing:
    - condition_var: if this evaluates to False, the command is not executed
    - assignment_var: name of a variable, to which the command output is assigned
    - command_type
    """

    return re.search(r'((.*)\?)?((.*)=)?(.*)', key).group(2, 4, 5)


class SystemsWrapper:
    def __init__(self, systems: dict):
        self.systems = systems

    def __getattr__(self, name):
        return self.systems[name].replace('hostname!', '')


class ConstantsWrapper:
    def __init__(self, constants: dict):
        self.constants = constants

    def __getattr__(self, name):
        return self.constants[name]


class AbortException(Exception):
    def __init__(self, return_code: int):
        self.return_code = return_code

    def __int__(self):
        return self.return_code


class SkipBatchItemException(Exception):
    pass


class UnknownCommandException(Exception):
    pass


class ReloadFromFile(Exception):
    def __init__(self, index: int):
        self.index = index

    def __int__(self):
        return self.index



from bundlewrap.exceptions import NoSuchNode, NoSuchGroup
from bundlewrap.group import Group
from bundlewrap.node import Node
from bundlewrap.repo import Repository

from .command import Command


class AutomatixBwRepo(Repository):
    def reload(self):
        self.__init__(repo_path=self.path)


class BWCommand(Command):
    def _generate_python_vars(self):
        locale_vars = {'AUTOMATIX_BW_REPO': self.env.config['bw_repo']}
        for key, value in self.env.systems.items():
            if not value.startswith('hostname!'):
                try:
                    self.env.config['bw_repo'].get_node(value)
                except NoSuchNode:
                    try:
                        self.env.config['bw_repo'].get_group(value)
                    except NoSuchGroup:
                        self.env.LOG.warning(f'"{value}" is neither a BW node nor a BW group')
        locale_vars['vars'] = self.env.vars
        locale_vars['NODES'] = BWNodesWrapper(repo=self.env.config['bw_repo'], systems=self.env.systems)
        return locale_vars

    def _remote_action(self) -> int:
        bw_repo: Repository = self.env.config['bw_repo']
        system = self.get_system()
        if system.startswith('hostname!'):
            return self._remote_action_on_hostname(hostname=system.replace('hostname!', ''))
        try:
            node: Node = bw_repo.get_node(system)
            return self._remote_action_on_hostname(hostname=node.hostname)
        except NoSuchNode as exc:
            try:
                group: Group = bw_repo.get_group(system)
            except NoSuchGroup:
                raise exc
            print()
            self.env.LOG.info(f' --- Executing command for all nodes in BW group >{group.name}< ---')
            for node in group.nodes:
                print()
                self.env.LOG.info(f'- {node.name} -')
                self._remote_bw_group_action(node=node)
            return 0

    def _remote_bw_group_action(self, node: Node):
        return_code = self._remote_action_on_hostname(hostname=node.hostname)
        if return_code != 0:
            self.env.LOG.error(f'Command ({self.index}) on {node.name} failed with return code {return_code}.')
            if self.env.cmd_args.force:
                return

            err_answer = self._ask_user(question='[PF] What should I do?', allowed_options=['p', 'T', 'v', 'r', 'a'])
            # answers 'a' and 'c' are handled by _ask_user, 'p' means just pass
            if err_answer == 'r':
                return self._remote_bw_group_action(node=node)


class BWNodesWrapper:
    def __init__(self, repo: Repository, systems: dict):
        self.repo = repo
        self.systems = systems

    def __getattr__(self, name):
        return self.repo.get_node(self.systems[name])



import subprocess
from argparse import Namespace
from os.path import abspath, dirname
from time import sleep

import pytest

from automatix import get_script, collect_vars, CONFIG, Command, SCRIPT_FIELDS
from automatix.automatix import Automatix
from automatix.logger import init_logger

SELFDIR = dirname(abspath(__file__))

default_args = Namespace(
    scriptfile=f'{SELFDIR}/test.yaml',
    systems=None,
    vars=None,
    secrets=None,
    print_overview=False,
    jump_to=0,
    interactive=False,
    force=False,
    debug=False,
)

script = get_script(args=default_args)

variables = collect_vars(script=script)

testauto = Automatix(
    script=script,
    variables=variables,
    config=CONFIG,
    cmd_class=Command,
    script_fields=SCRIPT_FIELDS,
    cmd_args=default_args,
    batch_index=1,
)

testauto.env.attach_logger()

environment = testauto.env

init_logger(name=CONFIG['logger'], debug=True)


def run_command_and_check(cmd):
    subprocess.run(cmd, shell=True).check_returncode()


@pytest.fixture(scope='function')
def ssh_up(docker_services):
    max_retries = 20
    for _ in range(max_retries):
        sleep(1)
        try:
            run_command_and_check(
                'ssh 2>/dev/null'
                ' -o StrictHostKeyChecking=no'
                ' -o "UserKnownHostsFile /dev/null"'
                ' -o ControlMaster=no'
                ' -o ControlPath=none'
                ' -o ConnectTimeout=10'
                ' docker-test /bin/true')
        except subprocess.CalledProcessError:
            continue
        run_command_and_check(cmd='ssh-keygen -R [localhost]:2222 >/dev/null 2>&1')
        run_command_and_check(cmd='ssh-keygen -R localhost:2222 >/dev/null 2>&1')
        run_command_and_check(cmd='ssh-keyscan -t ecdsa -p 2222 localhost 2>/dev/null >> ~/.ssh/known_hosts')
        return
    raise Exception('Maximum retries exceeded: SSH test setup could not be created.')



import locale
import subprocess


def _call(*args, **kwargs):
    try:
        return subprocess.check_output(*args, **kwargs).decode(locale.getpreferredencoding()).splitlines()
    except subprocess.CalledProcessError:
        return []


class ScriptFileCompleter(object):
    """
    Scriptfile completer
    """

    def __init__(self, script_path: str):
        self.script_path = script_path

    def __call__(self, prefix, **kwargs):
        completion = []
        pre_len = len(self.script_path) + 1
        directories = _call(["bash", "-c", f"compgen -A directory -- '{self.script_path}/{prefix}'"])
        for directory in directories:
            for ext in ['yaml', 'yml']:
                files = _call(["bash", "-c", f"compgen -A file -X '!*.{ext}' -- '{directory}/'"])
                completion += [f[pre_len:] for f in files]
        for ext in ['yaml', 'yml']:
            files = _call(["bash", "-c", f"compgen -A file -X '!*.{ext}' -- '{self.script_path}/{prefix}'"])
            completion += [f[pre_len:] for f in files]
        return completion



from unittest import TestCase

from automatix.config import _overwrite, _tupelize, check_deprecated_syntax

tc = TestCase()


def test__overwrite():
    start_dict = {
        'variables': {
            'a': 7,
            'b': 'test',
            'c': True,
        }
    }

    _overwrite(script=start_dict, key='variables', data=['a=27', 'b=true', 'd=s8'])

    assert start_dict == {
        'variables': {
            'a': '27',
            'b': 'true',
            'c': True,
            'd': 's8',
        }
    }


def test__tupelize():
    assert _tupelize('1.2.3.4') == (1, 2, 3, 4)
    with tc.assertRaises(ValueError):
        _tupelize('4.7.dev32')


def test__check_deprecated_syntax__dict(caplog):
    check_deprecated_syntax(ckey='local', entry={'test': ''}, script={}, prefix='')
    assert 'Command is not a string' in caplog.text


def test__check_deprecated_syntax__const(caplog):
    check_deprecated_syntax(ckey='local', entry='{const_apt}', script={}, prefix='')
    assert 'Using "{const_apt}" is deprecated. Use "{CONST.apt}" instead.' in caplog.text


def test__check_deprecated_syntax__global(caplog):
    check_deprecated_syntax(ckey='python', entry='import xy; global xy', script={}, prefix='[always:5]')
    assert '[always:5] Using "global xy" is deprecated. Use "PERSISTENT_VARS[\'xy\'] = xy" instead.'



import builtins
from copy import deepcopy
from subprocess import CalledProcessError
from unittest import mock

import pytest

from automatix.command import Command, parse_key
from tests.test_environment import environment, run_command_and_check, ssh_up  # noqa: F401


def test__integration__execute_remote_cmd(ssh_up):  # noqa: F811
    cmd = Command(
        cmd={'remote@testsystem': 'touch /test_remote_cmd'},
        index=2,
        pipeline='pipeline',
        env=environment,
        position=1,
    )
    cmd.execute()
    try:
        run_command_and_check('ssh docker-test ls /test_remote_cmd >/dev/null')
    except CalledProcessError:
        pytest.fail('Check for remote file not successful')


def test__execute_local_cmd(capfd):
    test_string = "Local Test String"

    # empty captured stdin and stderr
    _ = capfd.readouterr()

    cmd = Command(
        cmd={'local': f'echo {test_string}'},
        index=2,
        pipeline='pipeline',
        env=environment,
        position=1,
    )
    cmd.execute()

    out, err = capfd.readouterr()
    assert test_string in out
    assert err == ''


def test__execute_local_with_condition(capfd):
    env = deepcopy(environment)
    env.vars.update({
        'false_var': False,
        'true_var': True,
        'empty_var': '',
        'none_var': None,
        'example_string': 'example',
    })

    condition_tests = {
        'false_var': False,
        'true_var': True,
        'none_var': False,
        'empty_var': False,
        'example_string': True,
    }

    for condition_var, condition_out in condition_tests.items():
        # empty captured stdin and stderr
        _ = capfd.readouterr()

        cmd = Command(
            cmd={f'{condition_var}?local': 'pwd'},
            pipeline='pipeline',
            index=2,
            env=env,
            position=1,
        )
        cmd.execute()

        out, err = capfd.readouterr()
        assert condition_out == ('automatix' in out)
        assert err == ''


def test__execute_python_cmd():
    test_cmd = """
from uuid import uuid4
from pprint import pprint
PERSISTENT_VARS.update(locals())
"""

    cmd = Command(cmd={'python': test_cmd}, index=2, pipeline='pipeline', env=environment, position=1)
    cmd.execute()

    cmd = Command(cmd={'python': 'print(uuid4())'}, index=2, pipeline='pipeline', env=environment, position=1)
    cmd.execute()


def test__parse_key():
    assert parse_key('python') == (None, None, 'python')
    assert parse_key('remote@v1') == (None, None, 'remote@v1')
    assert parse_key('host=remote@v1') == (None, 'host', 'remote@v1')
    assert parse_key('is_jira?host=remote@v1') == ('is_jira', 'host', 'remote@v1')
    assert parse_key('is_jira!?python') == ('is_jira!', None, 'python')


def test__show_and_change_variables():
    cmd = Command(cmd={'python': 'pass'}, index=2, pipeline='pipeline', env=deepcopy(environment), position=1)
    assert cmd.env.vars == {
        'a': '{a}',
        'myvar': 'huhu',
        'cond': '{cond}',
        'cond2': '{cond2}',
    }
    with mock.patch.object(builtins, 'input', lambda _: 'var1=xyz'):
        cmd.show_and_change_variables()

    with mock.patch.object(builtins, 'input', lambda _: 'myvar=hallo'):
        cmd.show_and_change_variables()

    with mock.patch.object(builtins, 'input', lambda _: ' cond2  =  !dgkls=432 \n\t  '):
        cmd.show_and_change_variables()

    assert cmd.env.vars == {
        'a': '{a}',
        'myvar': 'hallo',
        'cond': '{cond}',
        'cond2': '!dgkls=432',
        'var1': 'xyz',
    }



BLUE = '\033[34m'
CYAN = '\033[36m'
GREEN = '\033[32m'
RED = '\033[31m'
YELLOW = '\033[33m'

BOLD = '\033[1m'
ITALIC = '\033[3m'

RESET = '\033[0m'


def blue(text):
    return f'{BLUE}{text}{RESET}'


def cyan(text):
    return f'{CYAN}{text}{RESET}'


def green(text):
    return f'{GREEN}{text}{RESET}'


def red(text):
    return f'{RED}{text}{RESET}'


def yellow(text):
    return f'{YELLOW}{text}{RESET}'


def bold(text):
    return f'{BOLD}{text}{RESET}'


def italic(text):
    return f'{ITALIC}{text}{RESET}'


def nocolor(text):
    return f'{RESET}{text}'



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='automatix_cmd',
    version='2.2.0',
    description='Automation wrapper for bash and python commands',
    keywords=['bash', 'shell', 'command', 'automation', 'process', 'wrapper', 'devops', 'system administration'],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/vanadinit/automatix_cmd',
    author='Johannes Paul',
    author_email='vanadinit@quantentunnel.de',
    license='MIT',
    python_requires='>=3.10',
    install_requires=[
        'pyyaml>=5.1',
    ],
    extras_require={
        'tests': ['cython<3.0.0', 'pytest', 'pytest-docker', 'flake8'],
        'bash completion': ['argcomplete'],
        'progress bar': ['python_progress_bar>=1.10'],
    },
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'automatix=automatix:main',
            'automatix-manager=automatix.parallel:run_manager',
            'automatix-from-file=automatix.parallel:run_auto_from_file',
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
)



import sys
from argparse import Namespace
from collections import OrderedDict
from typing import List

from .command import Command, AbortException, SkipBatchItemException, PERSISTENT_VARS, ReloadFromFile
from .config import get_script
from .environment import PipelineEnvironment


class Automatix:
    def __init__(
            self,
            script: dict,
            variables: dict,
            config: dict,
            cmd_class: type,
            script_fields: OrderedDict,
            cmd_args: Namespace,
            batch_index: int,
    ):
        self.script = script
        self.script_fields = script_fields
        self.cmd_class = cmd_class
        self.env = PipelineEnvironment(
            name=script['name'],
            config=config,
            systems=script.get('systems', {}),
            variables=variables,
            imports=script.get('imports', []),
            batch_mode=script.get('batch_mode', False),
            batch_items_count=script.get('batch_items_count', 1),
            batch_index=batch_index,
            cmd_args=cmd_args,
        )

        self._command_lists: dict = {}

    def command_list(self, pipeline: str) -> list:
        if pipeline == 'main':
            pipeline = 'pipeline'
        if not self._command_lists.get(pipeline):
            self._command_lists[pipeline] = self.build_command_list(pipeline=pipeline)
        return self._command_lists[pipeline]

    def get_command_position(self, index: int, pipeline: str) -> int:
        if pipeline == 'always':
            return index
        if pipeline in ['pipeline', 'main']:
            return index + len(self.command_list('always'))
        if pipeline == 'cleanup':
            return index + len(self.command_list('always')) + len(self.command_list('main'))

    def set_command_count(self):
        if not self.env.command_count:
            self.env.command_count = len(
                self.command_list('always') + self.command_list('main') + self.command_list('cleanup')
            )

    def build_command_list(self, pipeline: str) -> List[Command]:
        command_list = []
        for index, cmd in enumerate(self.script.get(pipeline, [])):
            new_cmd = self.cmd_class(
                cmd=cmd,
                index=index,
                env=self.env,
                pipeline=pipeline,
                position=self.get_command_position(index=index, pipeline=pipeline),
            )
            command_list.append(new_cmd)
            if new_cmd.assignment_var and new_cmd.assignment_var not in self.env.vars:
                self.env.vars[new_cmd.assignment_var] = f'{{{new_cmd.assignment_var}}}'
        return command_list

    def reload_script(self):
        self.script = get_script(args=self.env.cmd_args)
        self._command_lists = {}  # Clear cache

    def print_main_data(self):
        print('\n')
        self.env.LOG.info(' ------ Overview ------')
        for field_key, field_value in self.script_fields.items():
            print()
            self.env.LOG.info(f'{field_value}:')
            for key, value in self.script.get(field_key, {}).items():
                self.env.LOG.info(f" {key}: {value}")

    def print_command_line_steps(self, command_list: List[Command]):
        print()
        self.env.LOG.info('Commandline Steps:')
        for cmd in command_list:
            self.env.LOG.info(f"({cmd.index}) [{cmd.orig_key}]: {cmd.get_resolved_value()}")

    def _execute_command_list(self, name: str, start_index: int, treat_as_main: bool):
        try:
            steps = self.script.get('steps')
            for cmd in self.command_list(name)[start_index:]:
                if treat_as_main:
                    if steps and (self.script['exclude'] == (cmd.index in steps)):
                        # Case 1: exclude is True  and index is in steps => skip
                        # Case 2: exclude is False and index is in steps => execute
                        print()
                        self.env.LOG.notice(f'({cmd.index}) Not selected for execution: skip')
                        continue
                    cmd.execute(interactive=self.env.cmd_args.interactive, force=self.env.cmd_args.force)
                else:
                    cmd.execute()
        except ReloadFromFile as exc:
            print()
            self.env.LOG.info(f'Reload script from file and retry => ({exc.index})')
            self.reload_script()
            self._execute_command_list(name=name, start_index=exc.index, treat_as_main=treat_as_main)

    def execute_pipeline(self, name: str):
        if not self.command_list(name):
            return

        if name == 'main':
            treat_as_main = True
            start_index = self.env.cmd_args.jump_to
        else:
            treat_as_main = False
            start_index = 0

        print()
        self.env.LOG.info('------------------------------')
        self.env.LOG.info(f' --- Start {name.upper()} pipeline ---')

        self._execute_command_list(name=name, start_index=start_index, treat_as_main=treat_as_main)

        print()
        self.env.LOG.info(f' --- End {name.upper()} pipeline ---')
        self.env.LOG.info('------------------------------\n')

    def run(self):
        print('\n')
        self.env.LOG.info('//////////////////////////////////////////////////////////////////////')
        self.env.LOG.info(f"---- {self.script['name']} ----")
        self.env.LOG.info('//////////////////////////////////////////////////////////////////////')

        PERSISTENT_VARS.clear()

        self.execute_pipeline(name='always')

        self.print_main_data()
        self.print_command_line_steps(command_list=self.command_list('main'))
        if self.env.cmd_args.print_overview:
            sys.exit()

        try:
            self.execute_pipeline(name='main')
        except (AbortException, SkipBatchItemException):
            self.env.LOG.debug('Abort requested. Cleaning up.')
            self.execute_pipeline(name='cleanup')
            self.env.LOG.debug('Clean up done. Exiting.')
            raise

        self.execute_pipeline(name='cleanup')

        self.env.LOG.info('---------------------------------------------------------------')
        self.env.LOG.info('Automatix finished: Congratulations and have a N.I.C.E. day :-)')
        self.env.LOG.info('---------------------------------------------------------------')
